#include "alg/glsequence.hpp"

#include <iostream>
#include <cassert>
#include <math.h>
#include <stdexcept>
using namespace std;

GlSequence::GlSequence(const Sequence &s, 
                       boost::shared_ptr<AnnotationColors> cm) 
  : Sequence(s),
    color_mapper(cm),
    char_pix_per_world_unit(2.5)
{
  seq->setDrawable(default_drawable());
}

GlSequence::GlSequence(const GlSequence &s)
  : Sequence(s),
    color_mapper(s.color_mapper),
    char_pix_per_world_unit(s.char_pix_per_world_unit)
{
  seq->setDrawable(copy_drawable(s.seq->drawable()));
}

GlSequence::GlSequence(const GlSequence *s)
  : Sequence(s),
    color_mapper(s->color_mapper),
    char_pix_per_world_unit(s->char_pix_per_world_unit)
{
  seq->setDrawable(copy_drawable(s->seq->drawable()));
}

GlSequence::GlSequence(const GlSequenceRef s)
  : Sequence( (SequenceRef)s ),
    color_mapper(s->color_mapper),
    char_pix_per_world_unit(s->char_pix_per_world_unit)
{
  seq->setDrawable(copy_drawable(s->seq->drawable()));
}

GlSequence &GlSequence::operator=(const GlSequence & s)
{
  if (this != &s) {
    Sequence::operator=(s);
    seq->setDrawable(copy_drawable(s.seq->drawable()));
    color_mapper = s.color_mapper;
    assert(char_pix_per_world_unit == s.char_pix_per_world_unit);
  }
  return *this;
}

DrawableRef GlSequence::default_drawable()
{
  ColorRef c(new Color(0.0, 0.0, 0.0));
  DrawableRef d(new Drawable(0.0, 0.0, 1.0, 12.0, c));
  return d;
}

DrawableRef GlSequence::copy_drawable(DrawableRef old_d)
{
  ColorRef c(old_d->color());
  DrawableRef d(new Drawable(old_d));
  // use the same color
  d->setColor(c);
  return d;
}

void GlSequence::setX(float value)
{
  seq->drawable()->setX(value);
}

float GlSequence::x() const
{
  return seq->drawable()->x();
}

void GlSequence::setY(GLfloat value)
{
  seq->drawable()->setY(value);
}

float GlSequence::y() const
{
  return seq->drawable()->y();
}

float GlSequence::z() const
{
  return seq->drawable()->z();
}

float GlSequence::height() const
{
  return seq->drawable()->height();
}

GLfloat GlSequence::right() const
{
  return size()+x();
}

GLfloat GlSequence::size() const
{
  return Sequence::size();
}

Sequence::size_type GlSequence::leftbase(GLfloat left) const
{
  left = ceil(left - x());
  if (left < 0)
    return 0;
  else if (left > Sequence::size() )
    return Sequence::size();
  else
    return (Sequence::size_type)left;
}

Sequence::size_type GlSequence::rightbase(GLfloat right) const
{
  right = floor(right) - x();
  if (right > Sequence::size())
    return Sequence::size();
  else if ( right < 0) 
    return 0;
  else 
    return (Sequence::size_type)right;
}

Sequence::const_iterator 
GlSequence::region_begin(GLfloat left, GLfloat right) const
{
  if ( leftbase(left) > Sequence::size() or left > right )
    return Sequence::end();
  else
    return Sequence::begin() + leftbase(left);
}

Sequence::const_iterator 
GlSequence::region_end(GLfloat left, GLfloat right) const
{
  if ( rightbase(right) > Sequence::size() or left > right )
    return Sequence::end();
  else
    return Sequence::begin() + rightbase(right); 
}

GlSequence GlSequence::subseq(size_type start, size_type count) const
{
  GlSequence new_seq(*this);
  new_seq.seq = seq->subseq(start, count);
  // make sure our subseq has a drawable attached to it
  // perhaps we should figure out correct x,y,z,h coords
  DrawableRef d(default_drawable());
  // and default to our current color
  ColorRef c(color());
  d->setColor(c);
  new_seq.seq->setDrawable(d);
  copy_children(new_seq, start, count);
  
  return new_seq;
}

void GlSequence::setColor(ColorRef &c)
{
  seq->drawable()->setColor(c);
}

ColorRef GlSequence::color()
{
  return seq->drawable()->color();
}

const ColorRef GlSequence::color() const
{
  return seq->drawable()->color();
}

int GlSequence::get_viewport_width_in_pixels()
{
  GLint viewport[4];
  glGetIntegerv(GL_VIEWPORT, viewport);
  return viewport[3]; // grab the viewport width
}

GLfloat GlSequence::pixelWidth(GLfloat left, GLfloat right) const
{
  return pixelWidth(left, right, get_viewport_width_in_pixels());
}

GLfloat
GlSequence::pixelWidth(GLfloat left, GLfloat right, int vp_width) const
{
  return round((right-left)/vp_width);
}

bool GlSequence::is_sequence_renderable(GLfloat left, GLfloat right) const
{
  return is_sequence_renderable(left, right, get_viewport_width_in_pixels());
}

bool GlSequence::is_sequence_renderable(GLfloat left, 
                                        GLfloat right, 
                                        int viewport_width) const
{
  GLfloat world_width = right - left;
  GLfloat pixels_needed = (char_pix_per_world_unit * world_width);

  // if the number of pixels taken up by rendering the characters 
  // that'd show up in the current ortho width is less than the window
  // width we can actually draw something 
   return pixels_needed < viewport_width;
}


void GlSequence::draw(GLfloat left, GLfloat right) const
{
  if ( not is_sequence_renderable(left, right) ) {
    draw_track(left, right);
  } else {
    draw_sequence(left, right);
  }
  draw_annotations(left, right);
}

void GlSequence::draw_box(GLfloat world_left, GLfloat world_right,
                          GLfloat left, GLfloat right, 
                          GLfloat height, GLfloat z) const
{
  GLfloat pixel_width = pixelWidth(world_left, world_right);
  GLfloat offset = height/2.0;
  GLfloat top = y() + offset;
  GLfloat bottom = y() - offset;
  
  // make our box be at least 1 pixel
  if ((right-left) < pixel_width) {
    right = left + pixel_width;
  }
  glBegin(GL_QUADS);
    glVertex3f(left,  top,    z);
    glVertex3f(left,  bottom, z);
    glVertex3f(right, bottom, z);
    glVertex3f(right, top,    z);
  glEnd();
}

void GlSequence::draw_track(GLfloat left, GLfloat right) const
{
  glColor3fv(color()->get());
  // draw main sequence track
  draw_box(left, right, x(), x()+Sequence::size(), height(), 0.0);
}

void GlSequence::draw_annotations(GLfloat left, GLfloat right) const
{
  // draw annotations
  GLfloat annotation_z = z() + 10.0;
  const std::list<annot>& annots = Sequence::annotations();
  const std::list<motif>& motifs = Sequence::motifs();
  for (std::list<annot>::const_iterator annot_itor = annots.begin();
       annot_itor != annots.end();
       ++annot_itor)
  {
    glColor3f(0.0, 0.8, 0.0);
    draw_box(left, right, x()+annot_itor->begin, x()+annot_itor->end, 
             height(), annotation_z);
  }
  // if motifs?
  for (std::list<motif>::const_iterator motifs_itor = motifs.begin();
       motifs_itor != motifs.end();
       ++motifs_itor)
  {
    glColor3fv(color_mapper->lookup("motif", motifs_itor->sequence).get());
    draw_box(left, right, x()+motifs_itor->begin, x()+motifs_itor->end, 
             height(), annotation_z+1.0);
  }

}

// this way of drawing characters, came from the red open gl book
const int PT = 1;
const int STROKE = 2;
const int END =3;

typedef struct charpoint {
  GLfloat x, y;
  int type;
} CP;

CP Adata[] = {
  {0, -5, PT}, {2.5, 5, PT}, {5, -5, STROKE}, 
  {0.75, -2, PT}, {4.25, -2, END}
};

CP Tdata[] = {
  {2.5, -5, PT}, {2.5,5, STROKE}, {0, 5, PT}, {5, 5, END}
};

CP Gdata[] = {
  {5, 3, PT}, {3, 5, PT}, {2, 5, PT}, {0, 3, PT}, {0, -3, PT},
  {2, -5, PT}, {3, -5, PT}, {5, -3, STROKE}, 
  {2.5, -1, PT}, {5, -1,PT}, {5, -5, END}
};

CP Cdata[] = {
  {4.9, 3, PT}, {3, 5, PT}, {2, 5, PT}, {0, 3, PT}, {0, -3, PT},
  {2, -5, PT}, {3, -5, PT}, {5, -3, END}
};

CP Xdata[] = {{ 0, 5, PT}, {5, -5,STROKE},{0,-5,PT},{5, 5, END}};
CP Ndata[] = {{ 0, -5, PT}, {0, 5, PT}, {5, -5, PT}, {5, 5, END}};

//! the maximum width used for a charcter glyph
const int max_glyph_width = 5; // unit ( glyph_coord )

static void drawLetter(CP *l, GLfloat z)
{
  glBegin(GL_LINE_STRIP);
  while(1) {
    switch (l->type) {
      case PT:
        glVertex3f(l->x, l->y, z);
        break;
      case STROKE:
        glVertex3f(l->x, l->y, z);
        glEnd();
        glBegin(GL_LINE_STRIP);
        break;
      case END:
        glVertex3f(l->x, l->y, z);
        glEnd();
        return;
        break;
      default:
         throw runtime_error("data structure failure");
    }
    l++;
  }
}

void GlSequence::draw_sequence(GLfloat left, GLfloat right) const
{
  // FIXME: basically this needs to be greater than the number of annotations
  const GLfloat z = 30;
  glLineWidth(1);
  glColor3fv(color()->get());

  Sequence::const_iterator seq_itor = region_begin(left, right);
  Sequence::const_iterator seq_end = region_end(left, right);
  Sequence::size_type basepair = 0;
  const float bp_per_world = 1.0; //( world coord )
  const float glyph_x_scale = 0.125; // unit = ( world coord / glyph coord )
  // compute how much space there should be to either size of a letter
  const float glyph_margin = (bp_per_world - glyph_x_scale * max_glyph_width) 
                           / 2.0;  

  assert(seq_end - seq_itor >= 0);
  while(seq_itor != seq_end)
  {
    assert ( basepair < Sequence::size() );
    glPushMatrix();
    glTranslatef( x()+leftbase(left) + basepair + glyph_margin, y(), 1.0 );
    glScalef(glyph_x_scale, 1.0, 1.0);
    switch (*seq_itor) {
      case 'A': case 'a':
        drawLetter(Adata, z);
        break;
      case 'T': case 't':
        drawLetter(Tdata, z);
        break;
      case 'G': case 'g':
        drawLetter(Gdata, z);
        break;
      case 'C': case 'c':
        drawLetter(Cdata, z);
        break;
      case 'N': case 'n':
        drawLetter(Ndata, z);
        break;
      default:
        drawLetter(Xdata, z);
        break;
    }
    glPopMatrix();
    ++seq_itor;
    ++basepair;
  }
}

bool operator==(const GlSequence &left, const GlSequence &right)
{
  return ( (left.x() == right.x()) and
           (left.y() == right.y()) and
           (left.z() == right.z()) and
           (left.height() == right.height()) and
           (left.color() == right.color()));
}

