#define BOOST_AUTO_TEST_MAIN
#include <boost/test/auto_unit_test.hpp>
#include <boost/assign/std/vector.hpp>
using namespace boost::assign;

#include <vector>
using namespace std;

#include "alg/conserved_path.hpp"

//! write tests for our conserved path types (just to make sure they
//! don't change on us.
BOOST_AUTO_TEST_CASE ( conserved_path_simple ) 
{
  vector<int> path;
  path += 3,4,5,1; // magic from boost assign

  ConservedPath cp1;
  BOOST_CHECK_EQUAL( cp1.size(), 0 ); //empty
  BOOST_CHECK( cp1.begin() == cp1.end() );
  
  cp1.score = 18;
  cp1.track_indexes = path;
  BOOST_CHECK_EQUAL( cp1.size(), path.size() );

  ConservedPath cp2(20, 18, path);
  BOOST_CHECK_EQUAL (cp1, cp2);
  BOOST_CHECK_EQUAL( cp2.size(), path.size() );

  ConservedPath::iterator cp2_itor;
  vector<int>::iterator path_itor;
  for (cp2_itor = cp2.begin(), path_itor = path.begin();
       cp2_itor != cp2.end() && path_itor != path.end();
       ++cp2_itor, ++path_itor)
  {
    BOOST_CHECK_EQUAL( *cp2_itor, *path_itor );
  }

  ConservedPath cp3( cp2 );
  BOOST_CHECK_EQUAL( cp3.size(), cp2.size() );
  ConservedPath::iterator cp3_itor;
  for (cp2_itor = cp2.begin(), cp3_itor = cp3.begin();
       cp2_itor != cp2.end() && cp3_itor != cp3.end();
       ++cp2_itor, ++cp3_itor)
  {
    BOOST_CHECK_EQUAL( *cp2_itor, *cp3_itor );
  }
}

//! do vector like operations work?
BOOST_AUTO_TEST_CASE ( conserved_path_vector )
{
  ConservedPath cp;
  cp.score = 21.0;
  
  BOOST_CHECK_EQUAL( cp.size(), 0 );
  cp.push_back( 3 );
  BOOST_CHECK_EQUAL( cp.size(), 1);
  BOOST_CHECK_EQUAL( cp[0], 3 );
  cp.push_back( 4 );
  BOOST_CHECK_EQUAL( cp.size(), 2);
  BOOST_CHECK_EQUAL( cp[1], 4 );
  cp.pop_back();
  BOOST_CHECK_EQUAL( cp.size(), 1 );
}

// does the reverse compliment test work?
BOOST_AUTO_TEST_CASE ( conserved_path_reverse_compliment ) 
{
  vector<int> path;
  path += 3,4,-5,1; // magic from boost assign

  ConservedPath cp1(20, 10, path);
  vector<bool> reversed = cp1.reverseComplimented();

  BOOST_CHECK_EQUAL( path.size(), reversed.size());
  vector<int>::iterator path_i = path.begin();
  vector<bool>::iterator reversed_i = reversed.begin();
  for(; path_i != path.end() and reversed_i != reversed.end(); 
      ++path_i, ++reversed_i)
  {
    // if we're less than zero our reversed flag should be true
    BOOST_CHECK_EQUAL( *path_i < 0, *reversed_i);
  }
}

BOOST_AUTO_TEST_CASE ( extended_conserved_path_simple ) 
{
  vector<int> path;
  path += 3,4,5,1; // magic from boost assign

  ConservedPath ecp1;
  ecp1.window_size = 30;
  ecp1.score = 18;
  ecp1.track_indexes = path;
  BOOST_CHECK_EQUAL ( ecp1.size(), path.size() );

  ConservedPath ecp2(30, 18.0, path);
  BOOST_CHECK_EQUAL ( ecp1, ecp2 );
  BOOST_CHECK_EQUAL ( ecp2.size(), path.size() );

  ConservedPath ecp_ne(35, 20.0, path);
  BOOST_CHECK(ecp2 != ecp_ne);

  ConservedPath cp1;
  cp1.score = 18;
  cp1.track_indexes = path;

  ConservedPath ecp3(cp1);
  BOOST_CHECK_EQUAL( ecp2, ecp3 );
  BOOST_CHECK_EQUAL( ecp3.size(), path.size() );
}

//! Can we grow succesfully?
BOOST_AUTO_TEST_CASE ( extended_conserved_path_growth )
{
  vector<int> path_base;
  path_base += 3,4,5,1; // magic from boost assign
  vector<int> path_next;
  path_next += 4,5,6,2; 
  vector<int> path_next_reversed;
  path_next_reversed += 4,5,6,-2; 
  vector<int> path_not_next;
  path_not_next += 4,5,6,5; 
  const int window_size = 4;

  ConservedPath ecp_base(window_size, 20, path_base);
  ConservedPath cp_next(window_size, 25, path_next);
  ConservedPath cp_next_reversed(window_size, 25, path_next_reversed);
  ConservedPath cp_not_next(window_size, 22, path_not_next);

  BOOST_CHECK(ecp_base.nextTo(cp_next));
  BOOST_CHECK(!ecp_base.nextTo(cp_not_next));
  // FIXME: should something like this actually extend?
  BOOST_CHECK(!ecp_base.nextTo(cp_next_reversed));

  const int growth = 2;
  ecp_base.extend(growth);
  BOOST_CHECK_EQUAL( ecp_base.window_size, window_size + growth );
}

BOOST_AUTO_TEST_CASE ( extended_conserved_normalized ) 
{
  vector<int> path;
  path += 3,-14,5,-19; // magic from boost assign
  vector<int> normalized_path;
  normalized_path += 3,4, 5,9;

  ConservedPath ecp1;
  ecp1.window_size = 10;
  ecp1.score = 18;
  ecp1.track_indexes = path;
  BOOST_CHECK_EQUAL ( ecp1.size(), path.size() );

  vector<ConservedPath::path_element> normalized_index(ecp1.normalizedIndexes());
 
  for(int i=0; i != ecp1.size(); ++i)
  {
    BOOST_CHECK_EQUAL(normalized_index[i], normalized_path[i]);
  }
}

BOOST_AUTO_TEST_CASE( conserved_path_less_than )
{
  vector<int> p111;
  p111 += 1,1,1;
  vector<int> p112;
  p112 += 1,1,2;
  vector<int> p115;
  p115 += 1,1,5; 
  vector<int> p112a;
  p112a += 1,1,1;
  vector<int> p1111;
  p1111 += 1,1,1,1; 
  
  ConservedPath cp111(10, 1.0, p111);
  ConservedPath cp112(10, 1.0, p112);
  ConservedPath cp112a(10, 1.0, p112a);
  ConservedPath cp115(10, 1.0, p115);
  ConservedPath cp1111(10, 1.0, p1111);
  
  BOOST_CHECK_EQUAL( cp111 < cp112, true );
  BOOST_CHECK_EQUAL( cp111 < cp1111, true );
  BOOST_CHECK_EQUAL( cp1111 < cp111, false );
  BOOST_CHECK_EQUAL( cp112 < cp112, false );
  BOOST_CHECK_EQUAL( cp112 == cp112, true );
  BOOST_CHECK_EQUAL( cp112 < cp115, true );
}

BOOST_AUTO_TEST_CASE( conserved_path_sort )
{
  vector<int> p111;
  p111 += 1,1,1;
  vector<int> p112;
  p112 += 1,1,2;
  vector<int> p115;
  p115 += 1,1,5;
  vector<int> p121;
  p121 += 1,2,1; 

  ConservedPath cp111(10, 1.0, p111);
  ConservedPath cp112(10, 1.0, p112);
  ConservedPath cp115(10, 1.0, p115);
  ConservedPath cp121(10, 1.0, p121);
  
  list<ConservedPath> paths;
  paths.push_back(cp115);
  paths.push_back(cp121);
  paths.push_back(cp112);
  paths.push_back(cp111);
  
  paths.sort();
  BOOST_REQUIRE_EQUAL( paths.size(), 4 );
  list<ConservedPath>::const_iterator path_i = paths.begin();
  
  BOOST_CHECK_EQUAL( *path_i, cp111 ); ++path_i;
  BOOST_CHECK_EQUAL( *path_i, cp112 ); ++path_i;
  BOOST_CHECK_EQUAL( *path_i, cp115 ); ++path_i;
  BOOST_CHECK_EQUAL( *path_i, cp121 );
}
